SystemOrganization addCategory: #CodeImport!
SystemOrganization addCategory: #'CodeImport-Tests'!

!ClassCommentReader methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:13'!
importCodeFrom: aCodeImporter
	aCodeImporter importClassComment: self.! !

!ClassCommentReader methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:33'!
setClassComment: aClassComment
	self theClass theNonMetaClass
						classComment: aClassComment
						stamp: changeStamp! !

Object subclass: #ChunkImporter
	instanceVariableNames: 'readStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport'!

!ChunkImporter commentStamp: '<historical>' prior: 0!
I'm an object in charge of import streams in chunk format.  Chunk format is the one that separates pieces of code with the $!! character.

I handle doIts normally, but some special cases like class organization, class comment or methods are handled via a double dispatch (See my method extensions for that).

How to use me:

ChunkImporter readStream: '2+2!!' readStream.

or

ChunkImporter readString: '2+2!!'

or

ChunkImporter fileInFileNamed: 'something.st'

or

ChunkImporter fileIn: (FileStream readOnlyFileNamed: 'something.st')!

!ChunkImporter class methodsFor: 'fileIn' stamp: 'GuillermoPolito 5/5/2012 01:43'!
fileIn: aFileStream
	^self new
		file: aFileStream;
		evaluateChunks! !

!ChunkImporter class methodsFor: 'fileIn' stamp: 'GuillermoPolito 5/5/2012 00:08'!
fileInFileNamed: aFileName
	self fileIn: (FileStream readOnlyFileNamed: aFileName).
! !

!ChunkImporter class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 01:43'!
readStream: aReadStream
	^self new
		readStream: aReadStream;
		evaluateChunks! !

!ChunkImporter class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/5/2012 01:55'!
readString: aByteString 
	^self readStream: aByteString readStream! !

!ChunkImporter methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 01:17'!
checkForPreamble: chunk
	((chunk beginsWith: '"Change Set:') and: [ChangeSet current preambleString == nil])
		ifTrue: [ChangeSet current preambleString: chunk].
	((chunk beginsWith: '"Postscript:') and: [ChangeSet current postscriptString == nil])
		ifTrue: [ChangeSet current postscriptString: chunk].
							
! !

!ChunkImporter methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 01:44'!
evaluateChunks
	"Evaluates the expressions from the text in the file and answers the result of compilation"

	| value |
	[readStream atEnd] whileFalse: [
		readStream skipSeparators.
		value := self evaluateNextChunk.
		self skipStyleChunk.
	].
	self flushChangesFile.
	^value! !

!ChunkImporter methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 02:21'!
evaluateNextChunk
	| value isMetadata nextChunk |
	isMetadata := self isNextChunkMetaData.
	nextChunk := self nextChunk.
	isMetadata ifFalse: [ self checkForPreamble: nextChunk ].
	value := Compiler evaluate: nextChunk logged: isMetadata.
	isMetadata ifTrue: [ value := value importCodeFrom: self ].
	^value.! !

!ChunkImporter methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 01:39'!
file: aFileStream
	self readStream: aFileStream.
	self selectTextConverterForCode.! !

!ChunkImporter methodsFor: 'evaluating' stamp: 'GuillermoPolito 5/5/2012 01:12'!
flushChangesFile
	"Yes, we need to do this in order to flush the file, puaj"
	Smalltalk logChange: '----End fileIn----'.! !

!ChunkImporter methodsFor: 'handling-metadata' stamp: 'GuillermoPolito 5/5/2012 02:24'!
importCategorizer: aClassOrganizer 
	aClassOrganizer changeFromString: self nextChunk! !

!ChunkImporter methodsFor: 'handling-metadata' stamp: 'GuillermoPolito 5/5/2012 02:32'!
importClassCategory: aClassCategoryReader 
	| methodSource |
	
	self flag: #fix.  "It is the one compiling the methods. WHY???"
	
	[methodSource := self nextChunk.
	 methodSource notEmpty]
		whileTrue: [ aClassCategoryReader compileSourceCode: methodSource ]! !

!ChunkImporter methodsFor: 'handling-metadata' stamp: 'GuillermoPolito 5/5/2012 02:32'!
importClassComment: aClassCommentReader 
	aClassCommentReader setClassComment: self nextChunk! !

!ChunkImporter methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 01:44'!
isChunkEnd
	^readStream peek = nil or: [ self isTerminatorMark ]! !

!ChunkImporter methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 01:38'!
isNextChunkMetaData
	"If the next chunk starts with $!!, it will be an expression that after evaluation returns an object who knows how to read some metadata. 
	I.e. A chunk for comment reading should be something like:
	
	!!SomeClass commentStamp: '<historical>' prior: 0!!
	
	The object returned by the expression of that tag must understand #scanFrom:
	"
	
	^readStream peekFor: $!!! !

!ChunkImporter methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 01:38'!
isNextStyleChunk
	"Style chunks are between $] and $["
	^readStream peek == $]! !

!ChunkImporter methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 01:46'!
isTerminatorMark
	^readStream peek = $!!! !

!ChunkImporter methodsFor: 'reading' stamp: 'GuillermoPolito 5/5/2012 01:38'!
nextChunk
	| terminator out |
	terminator := $!!.
	out := (String new: 1000) writeStream.
	[ self isChunkEnd ] whileFalse: [
		out nextPut: readStream next.
	].
	self skipDoubleTerminatorMark.
	^out contents! !

!ChunkImporter methodsFor: 'accessing' stamp: 'GuillermoPolito 5/5/2012 01:39'!
readStream: aReadStream
	readStream := aReadStream.! !

!ChunkImporter methodsFor: 'reading' stamp: 'GuillermoPolito 5/5/2012 01:38'!
readUpToEndOfStyleChunk
	"Style chunks are between $] and $["
	^readStream upTo: $[! !

!ChunkImporter methodsFor: 'fileHandling' stamp: 'GuillermoPolito 5/5/2012 01:38'!
selectTextConverterForCode
	self flag: #fix.  "This should not be here probably."
	
	"We need to see the first three bytes in order to see the origin of the file"
	readStream binary.
	((readStream next: 3) = #[ 16rEF 16rBB 16rBF ]) ifTrue: [
		readStream converter: UTF8TextConverter new
	] ifFalse: [
		readStream converter: MacRomanTextConverter new.
	].

	"we restore the position to the start of the file again"
	readStream position: 0.
	
	"We put the file in text mode for the file in"
	readStream text.! !

!ChunkImporter methodsFor: 'skipping' stamp: 'GuillermoPolito 5/5/2012 01:38'!
skipDoubleTerminatorMark
	"if it comes a terminator make again, we skip it."
	"PRECONDITION: I must be called after checking that my previous character was a terminator.  I DO NOT CHECK IT."
	self isTerminatorMark ifTrue: [ readStream next. ].! !

!ChunkImporter methodsFor: 'skipping' stamp: 'GuillermoPolito 5/5/2012 01:38'!
skipStyleChunk
	"A style chunk is between $] and $[, so we skip everything in the middle"

	| pos |
	pos := readStream position.
	readStream skipSeparators.
	self isNextStyleChunk
		ifTrue: [ self readUpToEndOfStyleChunk ]
! !

!ClassCategoryReader methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:32'!
compileSourceCode: someSourceCode
	self theClass
			compile: someSourceCode 
			classified: category
			withStamp: changeStamp
			notifying: nil! !

!ClassCategoryReader methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:13'!
importCodeFrom: aCodeImporter
	aCodeImporter importClassCategory: self.! !

TestCase subclass: #ChunkImportTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CodeImport-Tests'!

!ChunkImportTestCase methodsFor: 'auxiliar' stamp: 'GuillermoPolito 5/5/2012 01:57'!
importAClass
	^ChunkImporter readStream: 'Object subclass: #CodeImportTestCaseTestClass
			instanceVariableNames: ''var1 var2 var3''
			classVariableNames: ''''
			poolDictionaries: ''''
			category: ''CodeImporter-Tests-Garbage''!!' readStream.! !

!ChunkImportTestCase methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 01:57'!
testImportAClass
	| class |
	SystemChangeNotifier uniqueInstance doSilently: [[
		class := self importAClass.
		self assert: #CodeImportTestCaseTestClass equals: class name.
		self assert: (class instVarNames includes: 'var1').
		self assert: (class instVarNames includes: 'var1').
		self assert: (class instVarNames includes: 'var1').
		self assert: class category equals: 'CodeImporter-Tests-Garbage'.
		
	] ensure: [ class ifNotNil: [ class removeFromSystem ] ] ]! !

!ChunkImportTestCase methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 02:20'!
testImportAClassCategory
	| class classOrganizationString chunk |
	classOrganizationString := '(''testing'' testImportAClass testImportAMethod)
(''auxiliar'' importAClass)'.
	chunk := '!!{1} reorganize!!'.
	SystemChangeNotifier uniqueInstance doSilently: [[
		class := self importAClass.
		
		"we write the methods we will categorize"
		class compileSilently: 'testImportAClass'.
		class compileSilently: 'testImportAMethod'.
		class compileSilently: 'importAClass'.
		
		ChunkImporter readString: (chunk format: { class name asString }), classOrganizationString.
		self assert: classOrganizationString trimBoth equals: class organization stringForFileOut trimBoth.
	] ensure: [ class ifNotNil: [ class removeFromSystem ] ] ]! !

!ChunkImportTestCase methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 02:01'!
testImportAClassComment
	| class comment |
	comment := 'I''m a nice comment, treat me well :).'.
	SystemChangeNotifier uniqueInstance doSilently: [[
		class := self importAClass.
		ChunkImporter readString: ('!!{1} commentStamp: ''<historical>'' prior: 0!!{2}!!' format: { class name asString . comment }).
		self assert: comment equals: class comment.
	] ensure: [ class ifNotNil: [ class removeFromSystem ] ] ]! !

!ChunkImportTestCase methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 02:03'!
testImportAMethod
	| class comment |
	SystemChangeNotifier uniqueInstance doSilently: [[
		class := self importAClass.
		ChunkImporter readString: ('!!{1} methodsFor: ''some protocol'' stamp: ''GuillermoPolito 5/2/2012 13:35''!!someMethod
	^true' format: { class name asString }).
	
		self assert: class new someMethod.
	] ensure: [ class ifNotNil: [ class removeFromSystem ] ] ]! !

!ChunkImportTestCase methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 01:41'!
testImportSimpleExpression
	self assert: 4 equals: (ChunkImporter readStream: '2+2!!' readStream)! !

!ChunkImportTestCase methodsFor: 'testing' stamp: 'GuillermoPolito 5/5/2012 01:55'!
testImportString
	self assert: 4 equals: (ChunkImporter readString: '2+2!!')! !

!Categorizer methodsFor: '*CodeImport' stamp: 'GuillermoPolito 5/5/2012 02:12'!
importCodeFrom: aCodeImporter
	aCodeImporter importCategorizer: self.! !
